/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* SCCS ID: @(#)resolve_dev.c	1.4 10/27/93 */
/*************************************************************************
 * FILE:      resolve_dev.c
 * Author:    Claude Saunders
 * Purpose:   Used in part to implement the devSend device access library
 *            calls. The device access library invokes the resolve_device_1()
 *            remote procedure call whenever a device (and message) is first
 *            referenced in a devSend.
 *
 * Mods:      login mm/dd/yy: description
 *************************************************************************/
#include <stdio.h>
#include <rpc/rpc.h>
#include <string.h>
#include <cadef.h>
#include <stdlib.h>
#include <string.h>
#include "dev.h"
#include "scan_defs.h"

/*************************************************************************
 * FUNCTION : resolve_device_1()
 * PURPOSE  : Given a device name and message (requesttype), return a 
 *            singly linked list of devicenode structs (see RPC interface
 *            definition file dev.x). The devicenode struct contains all
 *            information the server has on that device-name/message pair.
 *            If request->dname is an atomic device, there will only be one
 *            node which describes what action to take for request->mesg.
 *            If request->dname is a composite device, there will be a node
 *            for request->mesg for each constituent atomic device.
 *
 * ARGS in  : request: struct containing dname and mesg fields: both strings
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : resolve_result.myerrno indicates whether resolve_result.reply
 *            is valid or not. The linked list is actually the .reply field.
 *            This technique is shown in the RPC manuals.
 ************************************************************************/
resolve_result *resolve_device_1(request)
     requesttype *request;
{
  PDEV pdev;             /* device header node */
  PPVMSG ppvmsg;         /* ptr to linked list of supported messages */
  PMSG pmsg;             /* ptr to message specific information node */
  devicelist *dlptr;     /* for building up resolve_result.reply     */
  devicelist dl, prevdl;
  int nohandle;
  int dummytype;
  int dummysettype = 0;
  int dummydir;
  int dummysetdir = 0;
  static resolve_result res;   /* RPC return MUST be static ! */

  xdr_free(xdr_resolve_result,(char *)&res);     /* free previous reply */
  pdev = dev_find(request->dname);               /* find dev in hash table */
  res.reply = NULL;             /* initialize reply list */
  res.myerrno = OK;
  dlptr = &res.reply;           /* get addr of reply list */

  if (pdev == NULL)
    res.myerrno = NOSUCHDEV;

  while (pdev != NULL) {
    nohandle = 0;                                /* flag for msg handling */
    ppvmsg = pdev->pvmsgh;                       /* search for assoc msg */
    while (ppvmsg != NULL) {
      pmsg = ppvmsg->msgh;
      if (!strcmp(pmsg->name,request->mesg))
	break;
      ppvmsg = ppvmsg->next;
    }
    if (ppvmsg == NULL && pdev->type != COMPOSITE) {
      res.myerrno = NOSUCHMSG;
      nohandle = 1;
    }

    dl = (devicenode *)malloc(sizeof(devicenode)); /* malloc a devicenode */
    if (*dlptr == NULL)
      *dlptr = dl;                                 /* set reply to new node */
    else
      prevdl->next = dl;                           /* add to reply list */
    
    prevdl = dl;

    dl->next = NULL;                               /* begin filling out node */
    dl->dname = strdup(pdev->name);

    if (nohandle) {
      dl->type = DUMMY;     /* means this device doesn't handle message */
      dl->msg = strdup(request->mesg);     
      dl->pvname = strdup(" ");   /* filler so rpc doesn't choke */
      dl->def = strdup(" ");
      dl->nargs = 1;
      if (dummysetdir)
	dl->dir = dummydir;
      else
	dl->dir = READ;
      if (dummysettype)
	dl->datatype = dummytype;
      else
	dl->datatype = DBR_DOUBLE;
    }
    else {
      dl->type = pdev->type;
    
      switch (pdev->type) {
      case COMPOSITE:  {                  /* put filler info into this node  */
	dl->msg = strdup(request->mesg);  /* so rpc protocol doesn't choke  */
	dl->pvname = strdup(" ");
	dl->def = strdup(" ");
	break;
      }
      case ATOMIC: {
	dl->msg = strdup(pmsg->name);       
	dl->pvname = strdup(ppvmsg->pvname);
	dl->dir = pmsg->dir;
	dummysetdir = 1;
	dummydir = pmsg->dir;
	dl->nargs = pmsg->nargs;
	dl->datatype = pmsg->type;
	dummysettype = 1;
	dummytype = pmsg->type;
	dl->def = strdup(pmsg->def);
	break;
      }
      }
    }
    
    pdev = pdev->next;                   /* advance to next dev */
  }
  return(&res);
}

